@@ -10,9 +10,9 @@ import logging
 import typing
 import shutil
 import gc
-import threading
+from pathlib import Path
 from concurrent.futures import ThreadPoolExecutor, TimeoutError as FutureTimeoutError
-from itp_interface.tools.training_data import TrainingData
+from itp_interface.tools.training_data import TrainingData, DataLayoutFormat
 
 # Conditional Ray import
 try:
@@ -25,11 +25,12 @@ except ImportError:
     RayUtils = None
 from itp_interface.tools.coq_executor import CoqExecutor
 from itp_interface.tools.lean_cmd_executor import Lean3Executor
-from itp_interface.tools.lean4_sync_executor import Lean4SyncExecutor
+from itp_interface.lean.simple_lean4_sync_executor import SimpleLean4SyncExecutor
 from itp_interface.tools.isabelle_executor import IsabelleExecutor
 from itp_interface.tools.coq_local_data_generation_transform import LocalDataGenerationTransform as CoqLocalDataGenerationTransform
 from itp_interface.tools.lean_local_data_generation_transform import LocalDataGenerationTransform as LeanLocalDataGenerationTransform
 from itp_interface.tools.lean4_local_data_generation_transform import Local4DataGenerationTransform as Lean4LocalDataGenerationTransform
+from itp_interface.lean.lean4_local_data_extraction_transform import Local4DataExtractionTransform as Lean4LocalDataExtractionTransform
 from itp_interface.tools.isabelle_local_data_generation_transform import LocalDataGenerationTransform as IsabelleLocalDataGenerationTransform
 from itp_interface.tools.coq_training_data_generator import GenericTrainingDataGenerationTransform, TrainingDataGenerationType
 
@@ -136,7 +137,7 @@ class RunDataGenerationTransforms(object):
             search_lean_exec.__enter__()
             return search_lean_exec
         def _print_lean4_callback():
-            search_lean4_exec = Lean4SyncExecutor(project_path, None, file_path, use_human_readable_proof_context=use_human_readable, suppress_error_log=log_error)
+            search_lean4_exec = SimpleLean4SyncExecutor(project_path, None, file_path, use_human_readable_proof_context=use_human_readable, suppress_error_log=log_error)
             search_lean4_exec.__enter__()
             return search_lean4_exec
         def _print_isabelle_callback():
@@ -144,7 +145,11 @@ class RunDataGenerationTransforms(object):
             search_isabelle_exec = IsabelleExecutor(project_path, file_path, use_human_readable_proof_context=use_human_readable, suppress_error_log=log_error, port=port)
             search_isabelle_exec.__enter__()
             return search_isabelle_exec
-        if isinstance(transform, CoqLocalDataGenerationTransform) or isinstance(transform, LeanLocalDataGenerationTransform) or isinstance(transform, IsabelleLocalDataGenerationTransform) or isinstance(transform, Lean4LocalDataGenerationTransform):
+        if isinstance(transform, CoqLocalDataGenerationTransform) or \
+            isinstance(transform, LeanLocalDataGenerationTransform) or \
+            isinstance(transform, IsabelleLocalDataGenerationTransform) or \
+            isinstance(transform, Lean4LocalDataGenerationTransform) or \
+            isinstance(transform, Lean4LocalDataExtractionTransform):
             if isinstance(transform, IsabelleLocalDataGenerationTransform) and transform.ray_resource_pool is not None:
                 # This is a blocking call
                 port = ray.get(transform.ray_resource_pool.wait_and_acquire.remote(1))[0]
@@ -157,7 +162,9 @@ class RunDataGenerationTransforms(object):
                 elif isinstance(transform, IsabelleLocalDataGenerationTransform):
                     exec = IsabelleExecutor(project_path, file_path, use_human_readable_proof_context=use_human_readable, suppress_error_log=log_error, port=port)
                 elif isinstance(transform, Lean4LocalDataGenerationTransform):
-                    exec = Lean4SyncExecutor(project_path, None, file_path, use_human_readable_proof_context=use_human_readable, suppress_error_log=log_error)
+                    exec = SimpleLean4SyncExecutor(project_path, None, file_path, use_human_readable_proof_context=use_human_readable, suppress_error_log=log_error)
+                elif isinstance(transform, Lean4LocalDataExtractionTransform):
+                    exec = SimpleLean4SyncExecutor(project_path, None, file_path, use_human_readable_proof_context=use_human_readable, suppress_error_log=log_error)
                 else:
                     raise Exception("Unknown transform")
                 with exec:
@@ -170,6 +177,8 @@ class RunDataGenerationTransforms(object):
                         transform(training_data, project_id, exec, _print_isabelle_callback, theorems, other_args)
                     elif isinstance(transform, Lean4LocalDataGenerationTransform):
                         transform(training_data, project_id, exec, _print_lean4_callback, theorems, other_args)
+                    elif isinstance(transform, Lean4LocalDataExtractionTransform):
+                        transform(training_data, project_id, exec, _print_lean4_callback, theorems, other_args)
                     else:
                         raise Exception("Unknown transform")
             finally:
@@ -195,13 +204,18 @@ class RunDataGenerationTransforms(object):
         metadata.data_filename_suffix = RunDataGenerationTransforms.get_data_filename_suffix(transform)
         metadata.lemma_ref_filename_prefix = RunDataGenerationTransforms.get_lemma_ref_filename_prefix(transform)
         metadata.lemma_ref_filename_suffix = RunDataGenerationTransforms.get_lemma_ref_filename_suffix(transform)
+        if isinstance(transform, Lean4LocalDataExtractionTransform):
+            layout = DataLayoutFormat.DECLARATION_EXTRACTION
+        else:
+            layout = DataLayoutFormat.THEOREM_PROVING
         training_data = TrainingData(
             output_dir,
             RunDataGenerationTransforms.get_meta_file_name(transform),
             metadata,
             transform.max_parallelism,
             remove_from_store_after_loading=True,
-            logger=logger)
+            logger=logger,
+            layout=layout)
         return training_data
 
     @staticmethod
@@ -265,7 +279,10 @@ class RunDataGenerationTransforms(object):
         if self._use_ray:
             object_store_memory_in_gb = 100
             memory_in_gb = 5
-            ray_dashboard = RayUtils.init_ray(num_of_cpus=pool_size, object_store_memory_in_gb=object_store_memory_in_gb)
+            if not RayUtils.is_ray_initialized():
+                ray_dashboard = RayUtils.init_ray(num_of_cpus=pool_size, object_store_memory_in_gb=object_store_memory_in_gb, memory_in_gb=memory_in_gb)
+            else:
+                ray_dashboard = "Ray already initialized"
             self.logger.info(f"==============================>[{transform.name}] Ray initialized with {transform.max_parallelism} CPUs, Memory=({memory_in_gb} GiB, Object Memory = {object_store_memory_in_gb} GiB)<==============================")
             self.logger.info(f"Ray Context:\n {ray_dashboard}")
         else:
@@ -274,6 +291,7 @@ class RunDataGenerationTransforms(object):
         job_idx = 0
         project_names = list(projects.keys())
         project_names.sort()
+        file_index = 0
         for project in project_names:
             # Create temporary directory for each project
             proj_name = os.path.basename(project)
@@ -289,14 +307,31 @@ class RunDataGenerationTransforms(object):
                 some_files_processed = True
                 job_more_args = file_args.get(file_path, {})
                 # Create temporary directory for each file
-                full_file_path = os.path.join(project_path, file_path)
-                relative_file_path = file_path
-                relative_file_path = relative_file_path.replace("/", ".").replace(".v", "").replace(".lean", "").replace(".thy", "")
+                fp = Path(file_path)
+                pp = Path(project_path)
+                if fp.is_absolute():
+                    # then leave it as is
+                    full_file_path = file_path
+                else:
+                    # Check if the file path is relative to the project path
+                    fp = fp.resolve()
+                    pp = pp.resolve()
+                    # Now check if fp is relative to pp
+                    fp_is_rel_to_pp = fp.is_relative_to(pp)
+                    if fp_is_rel_to_pp:
+                        full_file_path = str(fp.relative_to(pp))
+                    else:
+                        # Just make it relative to project path
+                        full_file_path = os.path.join(project_path, file_path)
+                assert os.path.exists(full_file_path), f"File path {full_file_path} does not exist"
+                file_index += 1
+
+                relative_file_path = f"transformer_{file_index}"
                 temp_file_dir = os.path.join(temp_project_dir, relative_file_path)
                 os.makedirs(temp_file_dir, exist_ok=True)
                 log_file = os.path.join(self.logging_dir, f"{relative_file_path}.log")
                 theorems = projects[project][file_path]
-                if isinstance(transform, Lean4LocalDataGenerationTransform):
+                if isinstance(transform, Lean4LocalDataGenerationTransform) or isinstance(transform, Lean4LocalDataExtractionTransform):
                     # For every theorem we need to create a separate job
                     for _idx, theorem in enumerate(theorems):
                         log_file = os.path.join(self.logging_dir, f"{relative_file_path}-{_idx}.log")
@@ -317,13 +352,18 @@ class RunDataGenerationTransforms(object):
         final_training_meta.data_filename_suffix = RunDataGenerationTransforms.get_data_filename_suffix(transform)
         final_training_meta.lemma_ref_filename_prefix = RunDataGenerationTransforms.get_lemma_ref_filename_prefix(transform)
         final_training_meta.lemma_ref_filename_suffix = RunDataGenerationTransforms.get_lemma_ref_filename_suffix(transform)
+        if isinstance(transform, Lean4LocalDataExtractionTransform):
+            layout = DataLayoutFormat.DECLARATION_EXTRACTION
+        else:
+            layout = DataLayoutFormat.THEOREM_PROVING
         final_training_data = TrainingData(
             new_output_dir,
             RunDataGenerationTransforms.get_meta_file_name(transform),
             final_training_meta,
             transform.max_parallelism,
             remove_from_store_after_loading=True,
-            logger=self.logger)
+            logger=self.logger,
+            layout=layout)
         last_job_idx = 0
         tds = [None]*len(job_spec)
         num_theorems = 0
@@ -389,10 +429,13 @@ class RunDataGenerationTransforms(object):
         shutil.rmtree(temp_output_dir)
 
     def run_all_local_transforms(self, pool_size: int, projects: typing.Dict[str, typing.Dict[str, str]], use_human_readable: bool, new_output_dir: str, log_error: bool, other_args: typing.Dict[str, typing.Dict[str, dict]] = {}):
+        os.makedirs(new_output_dir, exist_ok=True)
         for idx, transform in enumerate(self.transforms):
             last_transform = idx == len(self.transforms) - 1
             save_transform = self.save_intermidiate_transforms or last_transform
-            self.run_local_transform(pool_size, transform, projects, use_human_readable, new_output_dir, log_error, save_transform, preserve_temp=self.save_intermidiate_transforms, other_args=other_args)
+            temp_new_output_dir = str(Path(new_output_dir) / str(idx))
+            os.makedirs(temp_new_output_dir, exist_ok=True)
+            self.run_local_transform(pool_size, transform, projects, use_human_readable, temp_new_output_dir, log_error, save_transform, preserve_temp=self.save_intermidiate_transforms, other_args=other_args)
         pass
 
 # Create Ray remote version if Ray is available
