@@ -6,15 +6,31 @@ if root_dir not in sys.path:
     sys.path.append(root_dir)
 import copy
 import os
-import jsonlines
-import typing
 import logging
 from dataclasses import dataclass, field
 from dataclasses_json import dataclass_json
 from collections import OrderedDict
-from typing import List, Optional, Tuple
+from typing import List, Optional, Union, runtime_checkable, Protocol
+from pydantic import BaseModel
+from itp_interface.lean.tactic_parser import FileDependencyAnalysis
 
-class MergableCollection(object):
+@runtime_checkable
+class TrainingDataFormat(Protocol):
+    
+    def to_json(self, indent=0) -> str:
+        raise NotImplementedError("to_json must be implemented by the child class")
+
+    @staticmethod
+    def load_from_file(file_path: str):
+        raise NotImplementedError("load_from_file must be implemented by the child class")
+    
+    @staticmethod
+    def load_from_string(json_text: str):
+        raise NotImplementedError("load_from_string must be implemented by the child class")
+
+
+@runtime_checkable
+class MergableCollection(Protocol):
     def merge(self, __o: object):
         raise NotImplementedError("merge must be implemented by the child class")
     
@@ -24,6 +40,32 @@ class MergableCollection(object):
     def __len__(self) -> int:
         raise NotImplementedError("__len__ must be implemented by the child class")
 
+
+@runtime_checkable
+class TrainingDataCollection(Protocol):
+    training_data: list
+
+    def merge(self, __o: object):
+        raise NotImplementedError("merge must be implemented by the child class")
+    
+    def undo_merge(self, size: int = 1, start_idx = 0) -> object:
+        raise NotImplementedError("undo_merge must be implemented by the child class")
+
+    def __len__(self) -> int:
+        return len(self.training_data)
+    
+    def to_json(self, indent=0) -> str:
+        raise NotImplementedError("to_json must be implemented by the child class")
+
+    @staticmethod
+    def load_from_file(file_path: str, logger: logging.Logger = None):
+        raise NotImplementedError("load_from_file must be implemented by the child class")
+
+    @staticmethod
+    def load_from_string(json_text: str, logger: logging.Logger = None):
+        raise NotImplementedError("load_from_string must be implemented by the child class")
+
+
 @dataclass_json
 @dataclass
 class LemmaRefWithScore(object):
@@ -81,6 +123,9 @@ class Goal(object):
     def __gt__(self, __o: object) -> bool:
         return self != __o and self >= __o
     
+    def to_json(self, indent=0) -> str:
+        return Goal.schema().dumps(self, indent=indent)
+
     @staticmethod
     def load_from_file(file_path: str):
         assert os.path.exists(file_path), "file_path must be a valid path to a file"
@@ -115,86 +160,31 @@ class LemmaReferences(object):
         return f"{self.lemma_name} : {self.lemma_defn}"
 #        return f"{self.lemma_defn} : {self.lemma_name}"
 
-    def clone(self, idx : typing.Optional[int] = None):
+    def clone(self, idx : Optional[int] = None):
         new_copy = copy.deepcopy(self)
         if idx is not None:
             new_copy.lemma_idx = idx
         return new_copy
-
-@dataclass_json
-@dataclass
-class LemmaReferencesCollection(MergableCollection):
-    """Class to store the lemma references."""
-    lemma_references: typing.List[LemmaReferences] = field(default_factory=list)
     
-    def __post_init__(self):
-        self._lemma_ref_to_idx = {lemma_ref: idx for idx, lemma_ref in enumerate(self.lemma_references)}
-
-    def merge(self, __o: object):
-        """
-        Merge the lemma references with another lemma references collection.
-        Returns the merged lemma references collection index map.
-        """
-        if not isinstance(__o, LemmaReferencesCollection) and not isinstance(__o, LemmaReferences) and not isinstance(__o, list):
-            raise TypeError(f"Cannot merge LemmaReferenceCollection with {type(__o)}")
-        if isinstance(__o, list) and not all(isinstance(x, LemmaReferences) for x in __o):
-            raise TypeError(f"Cannot merge LemmaReferenceCollection with list of {type(__o)}")
-        if isinstance(__o, LemmaReferences):
-            __o = [__o]
-        elif isinstance(__o, LemmaReferencesCollection):
-            __o = __o.lemma_references
-        to_take_cnt = len(__o)
-        new_idx_map = [-1] * to_take_cnt
-        for idx in range(to_take_cnt):
-            lemma_ref = __o[idx]
-            assert 0 <= lemma_ref.lemma_idx < len(__o), f"lemma_idx must be in range [0, {len(__o)}"
-            if lemma_ref not in self._lemma_ref_to_idx:
-                self._lemma_ref_to_idx[lemma_ref] = len(self.lemma_references)
-                lemma_ref_copy = copy.deepcopy(lemma_ref)
-                lemma_ref_copy.lemma_idx = len(self.lemma_references)
-                self.lemma_references.append(lemma_ref_copy)
-                new_idx_map[idx] = lemma_ref_copy.lemma_idx
-            else:
-                lemma_idx = self._lemma_ref_to_idx[lemma_ref]
-                new_idx_map[idx] = lemma_idx
-                self.lemma_references[lemma_idx].ref_count += lemma_ref.ref_count
-        assert all(idx != -1 for idx in new_idx_map), "new_idx_map must not contain any -1 values"
-        return new_idx_map
-    
-    def __len__(self) -> int:
-        return len(self.lemma_references)
-    
-    def __iter__(self):
-        return iter(self.lemma_references)
-    
-    def __getitem__(self, idx: int) -> LemmaReferences:
-        return self.lemma_references[idx]
+    def to_json(self, indent=0) -> str:
+        return LemmaReferences.schema().dumps(self, indent=indent)
     
     @staticmethod
-    def load_from_file(file_path: str, logger: logging.Logger = None):
-        assert os.path.exists(file_path), f"file_path:{file_path} must be a valid path to a file"
+    def load_from_file(file_path: str):
+        assert os.path.exists(file_path), "file_path must be a valid path to a file"
         json_text = None
-        if logger is not None:
-            logger.info(f"Loading json data from {file_path}")
         with open(file_path, "r") as f:
             json_text = f.read()
-        if logger is not None:
-            logger.info(f"Loaded json data from {file_path}")
-        return LemmaReferencesCollection.load_from_string(json_text, logger)
-
+        return LemmaReferences.load_from_string(json_text)
+    
     @staticmethod
-    def load_from_string(json_text: str, logger: logging.Logger = None):
+    def load_from_string(json_text: str):
         assert json_text is not None, "json_text cannot be None"
-        if logger is not None:
-            logger.info(f"Deseiralizing json data from string of length {len(json_text)} characters")
-        deserialized = LemmaReferencesCollection.schema().loads(json_text)
-        if logger is not None:
-            logger.info(f"Deseiralized json data from string of length {len(json_text)} characters")
-        return deserialized
+        return LemmaReferences.schema().loads(json_text)
 
 @dataclass_json
 @dataclass
-class TrainingDataFormat(object):
+class TheoremProvingTrainingDataFormat(object):
     """Class to format the training data for coq based automatic theorem provers.
     This class is responsible for formatting the training data for coq based automatic theorem provers.
     """
@@ -211,7 +201,7 @@ class TrainingDataFormat(object):
     theorem_name: Optional[str] = None # The name of the theorem.
 
     def __eq__(self, __o: object) -> bool:
-        if not isinstance(__o, TrainingDataFormat):
+        if not isinstance(__o, TheoremProvingTrainingDataFormat):
             return False
         goal_set_a = set([goal.goal for goal in self.start_goals])
         goal_set_b = set([goal.goal for goal in __o.start_goals])
@@ -251,7 +241,7 @@ class TrainingDataFormat(object):
     
     def __le__(self, __o: object) -> bool:
         # TrainingDataFormat 'a' is less (hard) than TrainingDataFormat 'b' iff all goals in 'a' are subset of goals in 'b'
-        if not isinstance(__o, TrainingDataFormat):
+        if not isinstance(__o, TheoremProvingTrainingDataFormat):
             raise TypeError(f"Cannot compare TrainingDataFormat with {type(__o)}")
         goal_set_a = set([goal.goal for goal in self.start_goals])
         goal_set_b = set([goal.goal for goal in __o.start_goals])
@@ -307,7 +297,7 @@ class TrainingDataFormat(object):
     
     def __ge__(self, __o: object) -> bool:
         # TrainingDataFormat 'a' is more (hard) than TrainingDataFormat 'b' iff all goals in 'b' are subset of goals in 'a'
-        if not isinstance(__o, TrainingDataFormat):
+        if not isinstance(__o, TheoremProvingTrainingDataFormat):
             raise TypeError(f"Cannot compare TrainingDataFormat with {type(__o)}")
         goal_set_a = set([goal.goal for goal in self.start_goals])
         goal_set_b = set([goal.goal for goal in __o.start_goals])
@@ -364,7 +354,7 @@ class TrainingDataFormat(object):
         return hash(tuple(goal_set))
 
     def have_same_proof_steps(self, __o: object) -> bool:
-        if not isinstance(__o, TrainingDataFormat):
+        if not isinstance(__o, TheoremProvingTrainingDataFormat):
             raise TypeError(f"Cannot compare TrainingDataFormat with {type(__o)}")
         return len(self.proof_steps) == len(__o.proof_steps) and all([p_a == p_b for p_a, p_b in zip(self.proof_steps, __o.proof_steps)])
 
@@ -377,23 +367,118 @@ class TrainingDataFormat(object):
 {hyps}
 """
 
+    def to_json(self, indent=0) -> str:
+        return TheoremProvingTrainingDataFormat.schema().dumps(self, indent=indent)
+
     @staticmethod
     def load_from_file(file_path: str):
         assert os.path.exists(file_path), "file_path must be a valid path to a file"
         json_text = None
         with open(file_path, "r") as f:
             json_text = f.read()
-        return TrainingDataFormat.load_from_string(json_text)
+        return TheoremProvingTrainingDataFormat.load_from_string(json_text)
     
     @staticmethod
     def load_from_string(json_text: str):
         assert json_text is not None, "json_text cannot be None"
-        return TrainingDataFormat.schema().loads(json_text)
+        return TheoremProvingTrainingDataFormat.schema().loads(json_text)    
 
 @dataclass_json
 @dataclass
-class TrainingDataCollection(MergableCollection):
-    training_data: List[TrainingDataFormat] = field(default_factory=list) # The list of training data.
+class LemmaReferencesCollection(TrainingDataCollection):
+    """Class to store the lemma references."""
+    training_data: list[LemmaReferences] = field(default_factory=list)
+    
+    def __post_init__(self):
+        self._lemma_ref_to_idx = {lemma_ref: idx for idx, lemma_ref in enumerate(self.training_data)}
+
+    def merge(self, __o: object):
+        """
+        Merge the lemma references with another lemma references collection.
+        Returns the merged lemma references collection index map.
+        """
+        if not isinstance(__o, LemmaReferencesCollection) and not isinstance(__o, LemmaReferences) and not isinstance(__o, list):
+            raise TypeError(f"Cannot merge LemmaReferenceCollection with {type(__o)}")
+        if isinstance(__o, list) and not all(isinstance(x, LemmaReferences) for x in __o):
+            raise TypeError(f"Cannot merge LemmaReferenceCollection with list of {type(__o)}")
+        if isinstance(__o, LemmaReferences):
+            __o = [__o]
+        elif isinstance(__o, LemmaReferencesCollection):
+            __o = __o.training_data
+        to_take_cnt = len(__o)
+        new_idx_map = [-1] * to_take_cnt
+        for idx in range(to_take_cnt):
+            lemma_ref = __o[idx]
+            assert 0 <= lemma_ref.lemma_idx < len(__o), f"lemma_idx must be in range [0, {len(__o)}"
+            if lemma_ref not in self._lemma_ref_to_idx:
+                self._lemma_ref_to_idx[lemma_ref] = len(self.training_data)
+                lemma_ref_copy = copy.deepcopy(lemma_ref)
+                lemma_ref_copy.lemma_idx = len(self.training_data)
+                self.training_data.append(lemma_ref_copy)
+                new_idx_map[idx] = lemma_ref_copy.lemma_idx
+            else:
+                lemma_idx = self._lemma_ref_to_idx[lemma_ref]
+                new_idx_map[idx] = lemma_idx
+                self.training_data[lemma_idx].ref_count += lemma_ref.ref_count
+        assert all(idx != -1 for idx in new_idx_map), "new_idx_map must not contain any -1 values"
+        return new_idx_map
+    
+    def undo_merge(self, size: int = 1, start_idx=0) -> object:
+        assert size >= 1, "size must be greater than equal to 1"
+        assert start_idx >= 0, "start_idx must be greater than zero"
+        assert start_idx < len(self.training_data), f"can only cut-down from idx < {len(self.training_data)}"
+        fraction = self.training_data[start_idx: size]
+        return LemmaReferencesCollection(training_data=fraction)
+    
+    def __iter__(self):
+        return iter(self.training_data)
+    
+    def __getitem__(self, idx: int) -> LemmaReferences:
+        return self.training_data[idx]
+
+    def to_json(self, indent=0) -> str:
+        return LemmaReferencesCollection.schema().dumps(self, indent=indent)
+
+    @staticmethod
+    def load_from_file(file_path: str, logger: logging.Logger = None):
+        assert os.path.exists(file_path), f"file_path:{file_path} must be a valid path to a file"
+        json_text = None
+        if logger is not None:
+            logger.info(f"Loading json data from {file_path}")
+        with open(file_path, "r") as f:
+            json_text = f.read()
+        if logger is not None:
+            logger.info(f"Loaded json data from {file_path}")
+        return LemmaReferencesCollection.load_from_string(json_text, logger)
+
+    @staticmethod
+    def load_from_string(json_text: str, logger: logging.Logger = None):
+        assert json_text is not None, "json_text cannot be None"
+        if logger is not None:
+            logger.info(f"Deserializing json data from string of length {len(json_text)} characters")
+        deserialized = LemmaReferencesCollection.schema().loads(json_text)
+        if logger is not None:
+            logger.info(f"Deserialized json data from string of length {len(json_text)} characters")
+        return deserialized
+
+@dataclass_json
+@dataclass
+class TheoremProvingTrainingDataCollection(TrainingDataCollection):
+    training_data: list[TheoremProvingTrainingDataFormat] = field(default_factory=list) # The list of training data.
+
+    def merge(self, __o: object):
+        assert isinstance(__o, TheoremProvingTrainingDataCollection)
+        self.training_data.extend(__o.training_data)
+    
+    def undo_merge(self, size: int = 1, start_idx=0) -> object:
+        assert size >= 1, "size must be greater than equal to 1"
+        assert start_idx >= 0, "start_idx must be greater than zero"
+        assert start_idx < len(self.training_data), f"can only cut-down from idx < {len(self.training_data)}"
+        fraction = self.training_data[start_idx: size]
+        return TheoremProvingTrainingDataCollection(training_data=fraction)
+
+    def to_json(self, indent=0) -> str:
+        return TheoremProvingTrainingDataCollection.schema().dumps(self, indent=indent)
 
     @staticmethod
     def load_from_file(file_path: str, logger: logging.Logger = None):
@@ -405,21 +490,54 @@ class TrainingDataCollection(MergableCollection):
             json_text = f.read()
         if logger is not None:
             logger.info(f"Loaded json data from {file_path}")
-        return TrainingDataCollection.load_from_string(json_text, logger)
+        return TheoremProvingTrainingDataCollection.load_from_string(json_text, logger)
 
     @staticmethod
     def load_from_string(json_text: str, logger: logging.Logger = None):
         assert json_text is not None, "json_text cannot be None"
         if logger is not None:
-            logger.info(f"Deseiralizing json data from string of length {len(json_text)} characters")
-        deserialized = TrainingDataCollection.schema().loads(json_text)
+            logger.info(f"Deserializing json data from string of length {len(json_text)} characters")
+        deserialized = TheoremProvingTrainingDataCollection.schema().loads(json_text)
         if logger is not None:
-            logger.info(f"Deseiralized json data from string of length {len(json_text)} characters")
+            logger.info(f"Deserialized json data from string of length {len(json_text)} characters")
         return deserialized
-    
+
+class ExtractionDataCollection(BaseModel):
+    training_data: list[FileDependencyAnalysis] = []
+
     def __len__(self) -> int:
         return len(self.training_data)
 
+    def to_json(self, indent=0) -> str:
+        if indent == 0:
+            return self.model_dump_json()
+        else:
+            return self.model_dump_json(indent=indent)
+
+    def merge(self, __o: object):
+        assert isinstance(__o, ExtractionDataCollection)
+        self.training_data.extend(__o.training_data)
+    
+    def undo_merge(self, size: int = 1, start_idx=0) -> object:
+        assert size >= 1, "size must be greater than equal to 1"
+        assert start_idx >= 0, "start_idx must be greater than zero"
+        assert start_idx < len(self.training_data), f"can only cut-down from idx < {len(self.training_data)}"
+        fraction = self.training_data[start_idx: size]
+        return ExtractionDataCollection(training_data=fraction)
+    
+    @staticmethod
+    def load_from_string(json_text: str, logger: logging.Logger = None):
+        assert json_text is not None, "json_text cannot be None"
+        return ExtractionDataCollection.model_validate_json(json_text)
+
+    @staticmethod
+    def load_from_file(file_path: str, logger: logging.Logger = None):
+        assert os.path.exists(file_path), "file_path must be a valid path to a file"
+        json_text = None
+        with open(file_path, "r") as f:
+            json_text = f.read()
+        return ExtractionDataCollection.load_from_string(json_text, logger=logger)
+
 @dataclass_json
 @dataclass
 class TrainingDataMetadataFormat(MergableCollection):
@@ -432,7 +550,7 @@ class TrainingDataMetadataFormat(MergableCollection):
     last_proof_id: Optional[str] = None
     external_theorems_used_cnt: int = 0
     local_theorems_used_cnt: int = 0
-    total_proof_step_cnt: int = 0
+    total_data_count: int = 0
     data_filename_prefix: str = "full_data"
     data_filename_suffix: str = ".json"
     lemma_ref_filename_prefix: str = "full_data_lemma_ref"
@@ -445,15 +563,20 @@ class TrainingDataMetadataFormat(MergableCollection):
         self.training_data_buffer_size = max(__o.training_data_buffer_size, self.training_data_buffer_size)
         self.last_training_data = __o.last_training_data
         self.last_proof_id = __o.last_proof_id
-        self.total_proof_step_cnt += __o.total_proof_step_cnt
+        self.total_data_count += __o.total_data_count
         self.external_theorems_used_cnt += __o.external_theorems_used_cnt
         self.local_theorems_used_cnt += __o.local_theorems_used_cnt
         self.num_theorems += __o.num_theorems
-
+    
+    def undo_merge(self, size: int = 1, start_idx=0) -> object:
+        raise NotImplementedError("undo_merge is not implemented for TrainingDataMetadataFormat")
     
     def __len__(self) -> int:
         return 0
 
+    def to_json(self, indent=0) -> str:
+        return TrainingDataMetadataFormat.schema().dumps(self, indent=indent)
+
     @staticmethod
     def load_from_file(file_path: str):
         assert os.path.exists(file_path), "file_path must be a valid path to a file"
@@ -476,15 +599,15 @@ class TrainingDataFormatLayout(object):
     def get_layout_format_name(self) -> str:
         raise NotImplementedError("get_layout_format_name must be implemented in derived classes")
 
-    def layout_training_data(self, training_data_format: TrainingDataFormat) -> typing.Union[str, typing.Tuple[str, str]]:
+    def layout_training_data(self, training_data_format: TheoremProvingTrainingDataFormat) -> Union[str, tuple[str, str]]:
         raise NotImplementedError("get_formatted_training_data must be implemented in derived classes")
     
-    def get_training_data_from_layout(self, formatted_training_data: str) -> TrainingDataFormat:
+    def get_training_data_from_layout(self, formatted_training_data: str) -> TheoremProvingTrainingDataFormat:
         raise NotImplementedError("get_training_data_format must be implemented in derived classes")
     
 if __name__ == "__main__":
     # Test the training data collection
-    training_data_format1 = TrainingDataFormat(
+    training_data_format1 = TheoremProvingTrainingDataFormat(
         proof_id="proof_id",
         start_goals=[
             Goal(hypotheses=[], goal="forall e : expr, size (constant_fold e) <= size e"),
@@ -500,7 +623,7 @@ if __name__ == "__main__":
         addition_state_info={}
     )
 
-    training_data_format2 = TrainingDataFormat(
+    training_data_format2 = TheoremProvingTrainingDataFormat(
         proof_id="proof_id",
         start_goals=[
             Goal(hypotheses=[], goal="forall e : expr, size (constant_fold e) <= size e"),
@@ -516,7 +639,7 @@ if __name__ == "__main__":
         addition_state_info={}
     )
 
-    training_data_format3 = TrainingDataFormat(
+    training_data_format3 = TheoremProvingTrainingDataFormat(
         proof_id="proof_id",
         start_goals=[
             Goal(hypotheses=[], goal="forall e : expr, size (constant_fold e) <= size e"),
@@ -531,7 +654,7 @@ if __name__ == "__main__":
         addition_state_info={}
     )
 
-    training_data_format4 = TrainingDataFormat(
+    training_data_format4 = TheoremProvingTrainingDataFormat(
         proof_id="proof_id",
         start_goals=[
             Goal(hypotheses=[], goal="forall e : expr, size (constant_fold e) <= size e"),
@@ -546,7 +669,7 @@ if __name__ == "__main__":
         addition_state_info={}
     )
 
-    training_data_format5 = TrainingDataFormat(
+    training_data_format5 = TheoremProvingTrainingDataFormat(
         proof_id="proof_id",
         start_goals=[
             Goal(hypotheses=[], goal="forall e : expr, size (constant_fold e) <= size e"),
@@ -560,7 +683,7 @@ if __name__ == "__main__":
         addition_state_info={}
     )
 
-    training_data_format6 = TrainingDataFormat(
+    training_data_format6 = TheoremProvingTrainingDataFormat(
         proof_id="proof_id",
         start_goals=[
             Goal(hypotheses=[], goal="forall e : expr, size (constant_fold e) <= size e"),
@@ -573,7 +696,7 @@ if __name__ == "__main__":
         addition_state_info={}
     )
 
-    training_data_format7 = TrainingDataFormat(
+    training_data_format7 = TheoremProvingTrainingDataFormat(
         proof_id="proof_id",
         start_goals=[
             Goal(hypotheses=[], goal="forall e : expr, size (constant_fold e) <= size e"),
