@@ -1,6 +1,26 @@
 import unittest
 import os
 import subprocess
+try:
+    import ray
+    from itp_interface.tools.ray_utils import RayResourcePoolActor, TimedRayExec, RayUtils
+    HAS_RAY = True
+except ImportError:
+    HAS_RAY = False
+    ray = None
+    RayResourcePoolActor = None
+    TimedRayExec = None
+    RayUtils = None
+
+def pretty_print_file_contents(dir_path):
+    print(f"Printing all files in the directory: {dir_path}")
+    for f in os.listdir(dir_path):
+        file_path = os.path.join(dir_path, f)
+        if os.path.isfile(file_path):
+            print('-'*50)
+            print(f"Contents of {file_path}:")
+            with open(file_path, "r") as file:
+                print(file.read())
 
 class TestDataGen(unittest.TestCase):
     def test_proof_step_data_gen(self):
@@ -42,11 +62,18 @@ class TestDataGen(unittest.TestCase):
         # Print the directory contents
         last_dir_path = os.path.join(".log/data_generation/benchmark/simple_benchmark_lean", last_dir)
         print("Last Directory Contents:", os.listdir(last_dir_path))
-        train_data = os.path.join(last_dir_path, "train")
+        train_data = os.path.join(last_dir_path, "train", "0")
         list_files = os.listdir(train_data)
         print("Train Directory Contents:", list_files)
         data_files = [f for f in list_files if f.endswith(".json") and f.startswith("local_data_")]
         print("Data Files:", data_files)
+        if len(data_files) == 0:
+            # Print the last directory contents again
+            pretty_print_file_contents(last_dir_path)
+            print('='*50)
+            # Open all the files in the train directory and print their contents for debugging
+            pretty_print_file_contents(train_data)
+               
         assert len(data_files) == 1, f"No files found in the train directory. Expected one file. Found: {data_files}"
         print(data_files[0])
         data_gen_file = os.path.join(train_data, data_files[0])
@@ -58,4 +85,12 @@ def main():
     unittest.main()
 
 if __name__ == '__main__':
+    if HAS_RAY:
+        os.environ["RAY_OBJECT_STORE_ALLOW_SLOW_STORAGE"] = "1"
+        object_store_memory_in_gb = 0.15
+        memory_in_gb = 0.25
+        ray_dashboard = RayUtils.init_ray(
+            num_of_cpus=2, 
+            object_store_memory_in_gb=object_store_memory_in_gb, 
+            memory_in_gb=memory_in_gb)
     main()
\ No newline at end of file
